/*
 *
 * X DLX Simulator
 *
 * The DLX is a 32-bit RISC machine based on the MIPS 2000 architecture.
 * The architecture includes 32 32-bit general purpose registers and a 
 * program counter.
 *
 * Floating point registers and an interrupt register are not implemented.
 *
 * The address space is 4 Gbytes, big-endian byte-addressable using 32-bit
 * addresses. All instructions must be properly aligned.
 *
 *
 * Cameron Humphries
 * Department of Computer Science
 * University of Adelaide
 * 1992
 *
 */
 
#include <stdio.h>
#include <X11/X.h>
#include <X11/Intrinsic.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/Form.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>

#include "MemList.h"
#include "sim.h"

#define SHOW_OCTAL 1
#define SHOW_DECIMAL 2
#define SHOW_HEX 3

#define SHOW_MEM_OCTAL 1
#define SHOW_MEM_DECIMAL 2
#define SHOW_MEM_HEX 3

#define NO_SELECTION 0
#define CODE_SELECTED 1
#define DATA_SELECTED 2

int code_type = SHOW_DECIMAL;
int mem_type = SHOW_MEM_DECIMAL;
int current_code_selection = -1;

extern int num_kb;
extern char *dis();
void Change_Code_Label();

char **code_arr;

static void octal_val(),decimal_val(),hex_val(),code_select();

static void mem_octal(),mem_decimal(),mem_hex();

Widget code_view,code_list,code_label;

int mem_loc_select = -1;
int selection_type = NO_SELECTION;

extern Widget data_list;

/*
 * 
 */
void Create_Code_View(Widget form)
{
  Widget w,popup_w,mem_popup_w;
  static Arg label_args[] = {
    {XtNlabel, (XtArgVal)"CODE (address: UNS, arguments: DEC)"},
    {XtNfromVert, (XtArgVal)NULL},
    {XtNborderWidth, (XtArgVal)0},
  };
  static Arg view_args[] = {
    {XtNfromVert, (XtArgVal)NULL},
    {XtNwidth, (XtArgVal)300},
    {XtNheight, (XtArgVal)365},
    {XtNallowVert, (XtArgVal)True},
    {XtNallowHoriz, (XtArgVal)True},
    {XtNuseBottom, (XtArgVal)True},
  };
  static Arg list_args[] = {
    {XtNlist, (XtArgVal)NULL},
    {XtNdefaultColumns, (XtArgVal)0},
    {XtNforceColumns, (XtArgVal)TRUE},
    {XtNnumberStrings, (XtArgVal)0},
    {XtNverticalList, (XtArgVal)True},
    {XtNcolumnSpacing, (XtArgVal)50},
    {XtNcallback, (XtArgVal)NULL},
  };
  static Arg menu_entry_args[] = {
    {XtNlabel, (XtArgVal)NULL},
    {XtNcallback, (XtArgVal)NULL},
  };
  static XtCallbackRec octal_cbk[] = {
    {octal_val, NULL},
    {NULL,NULL},
  };
  static XtCallbackRec decimal_cbk[] = {
    {decimal_val, NULL},
    {NULL,NULL},
  };
  static XtCallbackRec hex_cbk[] = {
    {hex_val, NULL},
    {NULL,NULL},
  };
  static XtCallbackRec mem_octal_cbk[] = {
    {mem_octal, NULL},
    {NULL,NULL},
  };
  static XtCallbackRec mem_decimal_cbk[] = {
    {mem_decimal, NULL},
    {NULL,NULL},
  };
  static XtCallbackRec mem_hex_cbk[] = {
    {mem_hex, NULL},
    {NULL,NULL},
  };
  static XtCallbackRec check_selection[] = {
    {code_select, NULL},
    {NULL,NULL},
  };
  static String translations = 
    "<Btn3Down>:XawPositionSimpleMenu(code_menu) MenuPopup(code_menu)";
    
  static String label_trans = 
    "<Btn3Down>:XawPositionSimpleMenu(code_mem_menu) MenuPopup(code_mem_menu)";
    
  /* Create popup representation menu */
  popup_w = XtCreatePopupShell("code_menu",simpleMenuWidgetClass,form,NULL,0);
  
  menu_entry_args[0].value = (XtArgVal)"OCTAL";
  menu_entry_args[1].value = (XtArgVal)octal_cbk;
  w = XtCreateManagedWidget("octal",smeBSBObjectClass,popup_w,
			    menu_entry_args,XtNumber(menu_entry_args));
			    
  menu_entry_args[0].value = (XtArgVal)"DECIMAL";
  menu_entry_args[1].value = (XtArgVal)decimal_cbk;
  w = XtCreateManagedWidget("decimal",smeBSBObjectClass,popup_w,
			    menu_entry_args,XtNumber(menu_entry_args));
			    
  menu_entry_args[0].value = (XtArgVal)"HEX";
  menu_entry_args[1].value = (XtArgVal)hex_cbk;
  w = XtCreateManagedWidget("hex",smeBSBObjectClass,popup_w,
			    menu_entry_args,XtNumber(menu_entry_args));
			    

  /* Create popup memory representation menu */
  mem_popup_w = XtCreatePopupShell("code_mem_menu",simpleMenuWidgetClass,form,
				   NULL,0);
  
  menu_entry_args[0].value = (XtArgVal)"OCTAL";
  menu_entry_args[1].value = (XtArgVal)mem_octal_cbk;
  w = XtCreateManagedWidget("octal",smeBSBObjectClass,mem_popup_w,
			    menu_entry_args,XtNumber(menu_entry_args));
			    
  menu_entry_args[0].value = (XtArgVal)"DECIMAL";
  menu_entry_args[1].value = (XtArgVal)mem_decimal_cbk;
  w = XtCreateManagedWidget("decimal",smeBSBObjectClass,mem_popup_w,
			    menu_entry_args,XtNumber(menu_entry_args));
			    
  menu_entry_args[0].value = (XtArgVal)"HEX";
  menu_entry_args[1].value = (XtArgVal)mem_hex_cbk;
  w = XtCreateManagedWidget("hex",smeBSBObjectClass,mem_popup_w,
			    menu_entry_args,XtNumber(menu_entry_args));
			    

  
  /* Create code view */
  label_args[1].value = (XtArgVal)XtNameToWidget(form,"File_Menu");
  code_label = XtCreateManagedWidget("Code_Title",labelWidgetClass,form,
				     label_args,XtNumber(label_args));
			    
  XtAugmentTranslations(code_label,XtParseTranslationTable(label_trans));

  view_args[0].value = (XtArgVal)code_label;
  code_view = XtCreateManagedWidget("Code_View",viewportWidgetClass,form,
			    view_args,XtNumber(view_args));
			    
  /* Create initial list */
  {
    int i,j;
    
    code_arr = (char **)malloc((num_kb*1024+1)*sizeof(char *));

    for (i=0;i<num_kb*1024;i++)
      {
	code_arr[i] = (String)malloc(60);
      }
    code_arr[num_kb*1024] = (String)malloc(60);
    *code_arr[num_kb*1024] = NULL;

    for (i=0,j=0;i<num_kb*1024;i+=4,j++)
      {
	if (i == regs[PC])
	  sprintf(code_arr[j],"%5u >>> %-30s",i,dis(imem(i),'d'));
	else
	  sprintf(code_arr[j],"%5u     %-30s",i,dis(imem(i),'d'));
	
      }
    *code_arr[num_kb*1024/4] = NULL;

  }
  
  list_args[0].value = (XtArgVal)code_arr;
  list_args[1].value = (XtArgVal)num_kb;
  list_args[3].value = (XtArgVal)(num_kb*1024/4);
  list_args[6].value = (XtArgVal)check_selection;
  code_list = XtCreateManagedWidget("Code_List",memListWidgetClass,code_view,
			    list_args,XtNumber(list_args));
			    
  XtAugmentTranslations(code_list,XtParseTranslationTable(translations));
  return;
}

static void octal_val(Widget w,XtPointer client_data,XtPointer call_data)
{
  int i,j;
  
  code_type = SHOW_OCTAL;
  for (i=0,j=0;i<num_kb*1024;i+=4,j++)
    {
      switch (mem_type)
	{
	case SHOW_MEM_HEX:
	  sprintf(code_arr[j],"%5x     %-30s",i,dis(imem(i),'o'));
	  break;
	case SHOW_MEM_DECIMAL:
	  sprintf(code_arr[j],"%5u     %-30s",i,dis(imem(i),'o'));
	  break;
	case SHOW_MEM_OCTAL:
	  sprintf(code_arr[j],"%5o     %-30s",i,dis(imem(i),'o'));
	  break;
	}
    }
  Change_Code_Label();
  XawMemListChange(code_list,code_arr,num_kb*1024/4,0,True);
  if (current_code_selection > -1)
    XawMemListHighlight(code_list,current_code_selection);
  return;
}

static void decimal_val(Widget w,XtPointer client_data,XtPointer call_data)
{
  int i,j;
  
  code_type = SHOW_DECIMAL;
  for (i=0,j=0;i<num_kb*1024;i+=4,j++)
    {
      switch (mem_type)
	{
	case SHOW_MEM_HEX:
	  sprintf(code_arr[j],"%5x     %-30s",i,dis(imem(i),'d'));
	  break;
	case SHOW_MEM_DECIMAL:
	  sprintf(code_arr[j],"%5u     %-30s",i,dis(imem(i),'d'));
	  break;
	case SHOW_MEM_OCTAL:
	  sprintf(code_arr[j],"%5o     %-30s",i,dis(imem(i),'d'));
	  break;
	}
    }
  Change_Code_Label();
  XawMemListChange(code_list,code_arr,num_kb*1024/4,0,True);
  if (current_code_selection > -1)
    XawMemListHighlight(code_list,current_code_selection);
  return;
}

static void hex_val(Widget w,XtPointer client_data,XtPointer call_data)
{
  int i,j;

  code_type = SHOW_HEX;
  for (i=0,j=0;i<num_kb*1024;i+=4,j++)
    {
      switch (mem_type)
	{
	case SHOW_MEM_HEX:
	  sprintf(code_arr[j],"%5x     %-30s",i,dis(imem(i),'x'));
	  break;
	case SHOW_MEM_DECIMAL:
	  sprintf(code_arr[j],"%5u     %-30s",i,dis(imem(i),'x'));
	  break;
	case SHOW_MEM_OCTAL:
	  sprintf(code_arr[j],"%5o     %-30s",i,dis(imem(i),'x'));
	  break;
	}
    }
  Change_Code_Label();
  XawMemListChange(code_list,code_arr,num_kb*1024/4,0,True);
  if (current_code_selection > -1)
    XawMemListHighlight(code_list,current_code_selection);
  return;
}

static void mem_octal(Widget w,XtPointer client_data,XtPointer call_data)
{
  int i,j;
  
  mem_type = SHOW_MEM_OCTAL;
  for (i=0,j=0;i<num_kb*1024;i+=4,j++)
    {
      switch (code_type)
	{
	case SHOW_HEX:
	  sprintf(code_arr[j],"%5o     %-30s",i,dis(imem(i),'x'));
	  break;
	case SHOW_DECIMAL:
	  sprintf(code_arr[j],"%5o     %-30s",i,dis(imem(i),'d'));
	  break;
	case SHOW_OCTAL:
	  sprintf(code_arr[j],"%5o     %-30s",i,dis(imem(i),'o'));
	  break;
	}
    }
  Change_Code_Label();
  XawMemListChange(code_list,code_arr,num_kb*1024/4,0,True);
  if (current_code_selection > -1)
    XawMemListHighlight(code_list,current_code_selection);
  return;
}

static void mem_decimal(Widget w,XtPointer client_data,XtPointer call_data)
{
  int i,j;
  
  mem_type = SHOW_MEM_DECIMAL;
  for (i=0,j=0;i<num_kb*1024;i+=4,j++)
    {
      switch (code_type)
	{
	case SHOW_HEX:
	  sprintf(code_arr[j],"%5u     %-30s",i,dis(imem(i),'x'));
	  break;
	case SHOW_DECIMAL:
	  sprintf(code_arr[j],"%5u     %-30s",i,dis(imem(i),'d'));
	  break;
	case SHOW_OCTAL:
	  sprintf(code_arr[j],"%5u     %-30s",i,dis(imem(i),'o'));
	  break;
	}
    }
  Change_Code_Label();
  XawMemListChange(code_list,code_arr,num_kb*1024/4,0,True);
  if (current_code_selection > -1)
    XawMemListHighlight(code_list,current_code_selection);
  return;
}

static void mem_hex(Widget w,XtPointer client_data,XtPointer call_data)
{
  int i,j;

  mem_type = SHOW_MEM_HEX;
  for (i=0,j=0;i<num_kb*1024;i+=4,j++)
    {
      switch (code_type)
	{
	case SHOW_HEX:
	  sprintf(code_arr[j],"%5x     %-30s",i,dis(imem(i),'x'));
	  break;
	case SHOW_DECIMAL:
	  sprintf(code_arr[j],"%5x     %-30s",i,dis(imem(i),'d'));
	  break;
	case SHOW_OCTAL:
	  sprintf(code_arr[j],"%5x     %-30s",i,dis(imem(i),'o'));
	  break;
	}
    }
  Change_Code_Label();
  XawMemListChange(code_list,code_arr,num_kb*1024/4,0,True);
  if (current_code_selection > -1)
    XawMemListHighlight(code_list,current_code_selection);
  return;
}

static void code_select(Widget w,XtPointer client_data,XtPointer call_data)
{
  if (((XawMemListReturnStruct *)call_data)->list_index == XAW_LIST_NONE)
    {
      selection_type = NO_SELECTION;
      mem_loc_select = -1;
      return;
    }
    
  if (selection_type == NO_SELECTION)
    {
      selection_type = CODE_SELECTED;
      mem_loc_select = ((XawMemListReturnStruct *)call_data)->list_index;
      current_code_selection = mem_loc_select;
    }
  else if (selection_type == DATA_SELECTED)
    {
      selection_type = CODE_SELECTED;
      mem_loc_select = ((XawMemListReturnStruct *)call_data)->list_index;
      current_code_selection = mem_loc_select;
      return;
    }
  else
    {
      mem_loc_select = ((XawMemListReturnStruct *)call_data)->list_index;
      current_code_selection = mem_loc_select;
      return;
    }
    
  return;
}

void Change_Code_Label()
{
  char str[200];

  static Arg args[] = {
    {XtNlabel, (XtArgVal)NULL},
  };
  
  switch (code_type)
    {
    case SHOW_OCTAL:
      switch (mem_type)
	{
	case SHOW_MEM_HEX:
	  sprintf(str,"CODE (address: HEX, arguments: OCT)");
	  break;
	case SHOW_MEM_DECIMAL:
	  sprintf(str,"CODE (address: UNS, arguments: OCT)");
	  break;
	case SHOW_MEM_OCTAL:
	  sprintf(str,"CODE (address: OCT, arguments: OCT)");
	  break;
	}
      break;
    case SHOW_DECIMAL:
      switch (mem_type)
	{
	case SHOW_MEM_HEX:
	  sprintf(str,"CODE (address: HEX, arguments: DEC)");
	  break;
	case SHOW_MEM_DECIMAL:
	  sprintf(str,"CODE (address: UNS, arguments: DEC)");
	  break;
	case SHOW_MEM_OCTAL:
	  sprintf(str,"CODE (address: OCT, arguments: DEC)");
	  break;
	}
      break;
    case SHOW_HEX:
      switch (mem_type)
	{
	case SHOW_MEM_HEX:
	  sprintf(str,"CODE (address: HEX, arguments: HEX)");
	  break;
	case SHOW_MEM_DECIMAL:
	  sprintf(str,"CODE (address: UNS, arguments: HEX)");
	  break;
	case SHOW_MEM_OCTAL:
	  sprintf(str,"CODE (address: OCT, arguments: HEX)");
	  break;
	}
      break;
    }
    
  args[0].value = (XtArgVal)str;
  XtSetValues(code_label,args,XtNumber(args));

  return;
}

Update_Code()
{
  int i,j;
  Widget w;

  switch (code_type) 
    {
    case SHOW_HEX:
      for (i=0,j=0;i<num_kb*1024;i+=4,j++)
	{
	  switch (mem_type)
	    {
	    case SHOW_MEM_HEX:
	      if (i == regs[PC])
		sprintf(code_arr[j],"%5x >>> %-30s",i,dis(imem(i),'x'));
	      else
		sprintf(code_arr[j],"%5x     %-30s",i,dis(imem(i),'x'));
	      break;
	    case SHOW_MEM_DECIMAL:
	      if (i == regs[PC])
		sprintf(code_arr[j],"%5u >>> %-30s",i,dis(imem(i),'x'));
	      else
		sprintf(code_arr[j],"%5u     %-30s",i,dis(imem(i),'x'));
	      break;
	    case SHOW_MEM_OCTAL:
	      if (i == regs[PC])
		sprintf(code_arr[j],"%5o >>> %-30s",i,dis(imem(i),'x'));
	      else
		sprintf(code_arr[j],"%5o     %-30s",i,dis(imem(i),'x'));
	      break;
	    }
	}
      break;

    case SHOW_OCTAL:
      for (i=0,j=0;i<num_kb*1024;i+=4,j++)
	{
	  switch (mem_type)
	    {
	    case SHOW_MEM_HEX:
	      if (i == regs[PC])
		sprintf(code_arr[j],"%5x >>> %-30s",i,dis(imem(i),'o'));
	      else
		sprintf(code_arr[j],"%5x     %-30s",i,dis(imem(i),'o'));
	      break;
	    case SHOW_MEM_DECIMAL:
	      if (i == regs[PC])
		sprintf(code_arr[j],"%5u >>> %-30s",i,dis(imem(i),'o'));
	      else
		sprintf(code_arr[j],"%5u     %-30s",i,dis(imem(i),'o'));
	      break;
	    case SHOW_MEM_OCTAL:
	      if (i == regs[PC])
		sprintf(code_arr[j],"%5o >>> %-30s",i,dis(imem(i),'o'));
	      else
		sprintf(code_arr[j],"%5o     %-30s",i,dis(imem(i),'o'));
	      break;
	    }
	}
      break;

    case SHOW_DECIMAL:
      for (i=0,j=0;i<num_kb*1024;i+=4,j++)
	{
	  switch (mem_type)
	    {
	    case SHOW_MEM_HEX:
	      if (i == regs[PC])
		sprintf(code_arr[j],"%5x >>> %-30s",i,dis(imem(i),'d'));
	      else
		sprintf(code_arr[j],"%5x     %-30s",i,dis(imem(i),'d'));
	      break;
	    case SHOW_MEM_DECIMAL:
	      if (i == regs[PC])
		sprintf(code_arr[j],"%5u >>> %-30s",i,dis(imem(i),'d'));
	      else
		sprintf(code_arr[j],"%5u     %-30s",i,dis(imem(i),'d'));
	      break;
	    case SHOW_MEM_OCTAL:
	      if (i == regs[PC])
		sprintf(code_arr[j],"%5o >>> %-30s",i,dis(imem(i),'d'));
	      else
		sprintf(code_arr[j],"%5o     %-30s",i,dis(imem(i),'d'));
	      break;
	    }
	}
      break;
    }

  XawMemListChange(code_list,code_arr,num_kb*1024/4,0,True);

  if ((regs[PC] < num_kb*1024) && (regs[PC]>=0))
    CentreMemLocation(code_view,code_list,regs[PC],num_kb);
    
  if ((current_code_selection > -1) && (current_code_selection < num_kb*1024))
    XawMemListHighlight(code_list,current_code_selection);
  return;
}

